#include "bst.h"
#include <catch.h>
#include <string>
#include <utility>

using Map = Bst_map<std::string, int>;

TEST_CASE("empty")
{
    Map map;
    CHECK( map.empty() );
    CHECK( map.lookup("hello") == nullptr );
}

// Helper to safely dereference a pointer that might be null.
template <class T>
auto& deref(T ptr)
{
    if (ptr == nullptr) throw std::runtime_error("deref: nullptr");
    else return *ptr;
}

// Helper to safely compare a pointer to its expected deferenced
// value.
template <class T>
struct Derefs_to
{
    T value;
};

// Compares the object of a pointer to the value in the `Derefs_to`
// only if the pointer is non-null. Otherwise returns false.
template <class T, class U>
bool operator==(T const* actual, Derefs_to<U> const& expected)
{
    return actual && *actual == expected.value;
}

// Factory function for constructing `Derefs_to`s without explicit
// type arguments.
template <class T>
Derefs_to<T> derefs_to(T value)
{
    return Derefs_to<T>{value};
}

TEST_CASE("insert and lookup")
{
    Map map;
    CHECK( map.lookup("map") == nullptr );
    map.insert("map", 3);
    CHECK( map.lookup("map") == derefs_to(3) );
    CHECK( map.lookup("lookup") == nullptr );

    ++deref(map.lookup("map"));
    CHECK( map.lookup("map") == derefs_to(4) );

    map.insert("map", 90);
    CHECK( map.lookup("map") == derefs_to(90) );
}

TEST_CASE("more elements")
{
    Map map;

    map.insert("d", 100);
    map.insert("b", 98);
    map.insert("a", 97);
    map.insert("c", 99);
    map.insert("f", 102);
    map.insert("e", 101);
    map.insert("g", 103);

    CHECK( map.lookup("a") == derefs_to(97) );
    CHECK( map.lookup("b") == derefs_to(98) );
    CHECK( map.lookup("c") == derefs_to(99) );
    CHECK( map.lookup("d") == derefs_to(100) );
    CHECK( map.lookup("e") == derefs_to(101) );
    CHECK( map.lookup("f") == derefs_to(102) );
    CHECK( map.lookup("g") == derefs_to(103) );

    CHECK( map.lookup("a*") == nullptr );
    CHECK( map.lookup("b*") == nullptr );
    CHECK( map.lookup("c*") == nullptr );
    CHECK( map.lookup("d*") == nullptr );
    CHECK( map.lookup("e*") == nullptr );
    CHECK( map.lookup("f*") == nullptr );
    CHECK( map.lookup("g*") == nullptr );
}
