#include <catch.h>
#include <string>

template <class T>
T max(T a, T b)
{
    return a < b ? b : a;
}


TEST_CASE("max")
{
    CHECK( max(3, 5) == 5 );
    CHECK( max(6.4, 3.2) == 6.4 );
    CHECK( max<std::string>("a", "b") == "b" );

    struct Foo {};

////    Type error:
//    max(Foo{}, Foo{});
}

template <class T>
struct Posn
{
    T x;
    T y;

    double distance(Posn) const;

    Posn operator+(Posn) const;
};

template<class T>
double Posn<T>::distance(Posn that) const
{
    T dx = x - that.x;
    T dy = y - that.y;
    return std::sqrt(dx * dx + dy * dy);
}

template<class T>
Posn<T> Posn<T>::operator+(Posn that) const
{
    return {x + that.x, y + that.y};
}

TEST_CASE("Posn::distance")
{
    Posn<int> p1{0, 0};
    Posn<int> p2{3, 4};

    CHECK( p1.distance(p1) == 0 );
    CHECK( p1.distance(p2) == 5 );

    Posn<double> p3{0, 0};
    Posn<double> p4{3, 4};
    CHECK( p3.distance(p3) == 0 );
    CHECK( p3.distance(p4) == 5 );

//    // Type error:
//    CHECK( p1.distance(p3) == 0 );
}

TEST_CASE("Posn::operator+")
{
    Posn<int> p1{2, 3};
    Posn<int> p2{5, 7};
    Posn<int> p3 = p1 + p2;

    CHECK( p3.x == 7 );
    CHECK( p3.y == 10 );

    Posn<std::string> p4{"foo", "bar"};
    Posn<std::string> p5{"baz", "qux"};
    Posn<std::string> p6 = p4 + p5;

    CHECK( p6.x == "foobaz" );
    CHECK( p6.y == "barqux" );
}
